import numpy as np
from scipy.spatial import distance
from scipy.sparse import issparse, isspmatrix_csr, csr_matrix, spmatrix
    
def uclab(X, n, alpha, drop_start=1, drop_rate=0):
    N = X.shape[0]
    sample_index = np.full(n, -1, dtype=int)
    distances = np.full(N, -1)
    
    # Define function
    #int64 = np.int64
    
    # step 0
    initial_index = np.random.randint(N, size=1)
    sample_index[0] = initial_index
    distances = distance.cdist(np.expand_dims(X[sample_index[0],:], axis=0), X).flatten()
    distances[initial_index] = -1
    d_index = np.argwhere(distances != -1).flatten()
    distances[d_index] = np.power(1/distances[d_index], alpha)
    
    for i in np.arange(1,n):
        #print(i, end=' ')
        # drop large points
        if i == np.int64(drop_start*n) and drop_rate != 0:
            current_number = np.count_nonzero(distances!=-1)
            drop_number = np.int64(current_number*drop_rate)        
            left_number = current_number - drop_number
            more_number = n - i
            # if drop too many, then return current best
            if more_number > left_number:
                drop_number = current_number - more_number
                sort_index = np.argsort(distances) 
                drop_index = sort_index[-drop_number:]
                distances[drop_index] = -1
                keep_index = np.where(distances != -1)[0]
                sample_index[i:] = keep_index
                return sample_index      
            sort_index = np.argsort(distances)            
            drop_index = sort_index[-drop_number:]
            distances[drop_index] = -1
        closest_index = np.where(distances > 0, distances, np.inf).argmin()
        sample_index[i] = closest_index
        distances[closest_index] = -1
        d_index = np.argwhere(distances != -1).flatten()
        d = distance.cdist(np.expand_dims(X[sample_index[i],:], axis=0), X[d_index,:]).flatten()
        d = np.power(1/d, alpha)
        distances[d_index] += d     
    return sample_index


def uclab_split(X, n, alpha, drop_start=1, drop_rate=0, split=4):
    # random shuffle index
    #np.random.seed(seed)
    index = np.arange(X.shape[0])
    np.random.shuffle(index)
    X = X[index,]
    # subsample per split
    X_list = np.array_split(X, split, axis=0)
    n_list = np.array_split(np.arange(n), split, axis=0)
    sample_index_list = [uclab(d, len(n_split), alpha, drop_start, drop_rate) for n_split, d in zip(n_list, X_list)]    
    # adjust split index
    sample_index_list_adjust = []
    adjust = 0
    for i, l in enumerate(sample_index_list):
        if i > 0:
            adjust += len(X_list[i-1])
            l = l + adjust
        sample_index_list_adjust.append(l)
    sample_index = np.concatenate(sample_index_list_adjust)
    # get original index
    sample_index = index[sample_index]
    return sample_index


import numpy as np
import pandas as pd
import scanpy as sc
import sklearn as sk
from anndata import AnnData
from numbers import Number
import warnings
from typing import Union, Optional, Tuple, Collection, Sequence, Iterable
import scipy as sp
from scipy.spatial import distance
from scipy.sparse import issparse, isspmatrix_csr, csr_matrix, spmatrix
from .uclab import uclab, uclab_split

def scsampler(
    data: Union[AnnData, np.ndarray, spmatrix],
    fraction: Optional[float] = None,
    n_obs: Optional[int] = None,
    random_state: int = 0,
    copy: bool = False,
    obsm: Optional[str] = 'X_pca',
    random_split: Optional[int] = None,
) -> Optional[AnnData]:
    """\
    Subsample to a fraction of the number of observations. This function refers to the subsample function in scanpy. 
    Parameters
    ----------
    data
        The (annotated) data matrix of shape `n_obs` × `n_vars`.
        Rows correspond to cells and columns to genes.
    fraction
        Subsample to this `fraction` of the number of observations.
    n_obs
        Subsample to this number of observations.
    random_state
        Random seed to change subsampling.
    copy
        If an :class:`~anndata.AnnData` is passed,
        determines whether a copy is returned.
    Returns
    -------
    Returns `X[obs_indices], obs_indices` if data is array-like, otherwise
    subsamples the passed :class:`~anndata.AnnData` (`copy == False`) or
    returns a subsampled copy of it (`copy == True`).
    """
    ## Sample space
    X = data.obsm[obsm] if isinstance(data, AnnData) else data
    
    np.random.seed(random_state)
    old_n_obs = data.n_obs if isinstance(data, AnnData) else data.shape[0]
    old_n_vars = X.shape[1]
    
    if n_obs is not None:
        new_n_obs = n_obs
    elif fraction is not None:
        if fraction > 1 or fraction < 0:
            raise ValueError(f'`fraction` needs to be within [0, 1], not {fraction}')
        new_n_obs = int(fraction * old_n_obs)
    else:
        raise ValueError('Either pass `n_obs` or `fraction`.')
    
    ## Run uclab
    split = 1 if random_split is None else random_split
    obs_indices = uclab_split(X, new_n_obs, alpha = 4 * old_n_vars, drop_start=1, drop_rate=0, split=split)
    
    if isinstance(data, AnnData):
        if copy:
            return data[obs_indices].copy()
        else:
            data._inplace_subset_obs(obs_indices)
    else:
        X = data
        return X[obs_indices], obs_indices



import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
    name="scsampler", # Replace with your own username
    version="1.0.0",
    author="Dongyuan Song",
    author_email="dongyuansong@ucla.edu",
    description="A package for fast diversity-preserving subsampling of large-scale single-cell transcriptomic data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SONGDONGYUAN1994/scsampler",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
